#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__author__ = "Jinliang Ma"

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.network import Arc

from idaes.core import FlowsheetBlock
from idaes.models_extra.power_generation.unit_models.watertank import WaterTank
from idaes.models.control.controller import PIDController, ControllerType
import idaes.core.util.scaling as iscale

# TODO: Should have a test for this that does not depend on models_extra
from idaes.models_extra.power_generation.unit_models.helm import (
    HelmValve as WaterValve,
    HelmIsentropicCompressor as WaterPump,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.properties import iapws95
from idaes.core.util.dyn_utils import copy_values_at_time, copy_non_time_indexed_values
from idaes.core.solvers import get_solver
from idaes.models.properties.general_helmholtz import helmholtz_available

import pytest


def set_scaling_factors(m):
    """Set scaling factors for variables and expressions. These are used for
    variable scaling and used by the framework to scale constraints.

    Args:
        m: plant model to set scaling factors for.

    Returns:
        None
    """
    # Set scaling factors for boiler system
    iscale.set_scaling_factor(m.fs.tank.control_volume.energy_holdup, 1e-10)
    iscale.set_scaling_factor(m.fs.tank.control_volume.material_holdup, 1e-6)
    if m.dynamic:
        for t, c in m.fs.tank.control_volume.energy_accumulation_disc_eq.items():
            iscale.constraint_scaling_transform(c, 1e-6)

    # scaling factor for control valves
    for t in m.fs.time:
        iscale.set_scaling_factor(
            m.fs.valve.control_volume.properties_in[t].flow_mol, 0.001
        )

    # Calculate calculated scaling factors
    iscale.calculate_scaling_factors(m)


@pytest.fixture
def m():
    m_ss = get_model(dynamic=False)
    m_dyn = get_model(dynamic=True)
    copy_non_time_indexed_values(m_dyn.fs, m_ss.fs, copy_fixed=True)
    for t in m_dyn.fs.time:
        copy_values_at_time(m_dyn.fs, m_ss.fs, t, 0.0, copy_fixed=True)
    m_dyn.fs.controller.mv_ref.value = m_dyn.fs.valve.valve_opening[0].value
    # calculate mv integral component assuming error is zero
    m_dyn.fs.controller.mv_integral_component[:].value = 0

    dof = degrees_of_freedom(m_dyn)
    assert dof == 0
    run_dynamic(m_dyn)
    return m_dyn


def get_model(dynamic=False):
    m = pyo.ConcreteModel(name="Testing PID controller model")
    if dynamic:
        m.dynamic = True
        m.fs = FlowsheetBlock(
            dynamic=True, time_set=[0, 50, 1000], time_units=pyo.units.s
        )
    else:
        m.dynamic = False
        m.fs = FlowsheetBlock(dynamic=False)
    m.fs.prop_water = iapws95.Iapws95ParameterBlock()

    # water pump
    m.fs.pump = WaterPump(dynamic=False, property_package=m.fs.prop_water)

    # water tank
    m.fs.tank = WaterTank(
        tank_type="simple_tank", has_holdup=True, property_package=m.fs.prop_water
    )

    m.fs.valve = WaterValve(
        dynamic=False, has_holdup=False, phase="Liq", property_package=m.fs.prop_water
    )

    if dynamic:
        m.fs.controller = PIDController(
            process_var=m.fs.tank.tank_level,
            manipulated_var=m.fs.valve.valve_opening,
            controller_type=ControllerType.PI,
        )

        m.discretizer = pyo.TransformationFactory("dae.finite_difference")
        m.discretizer.apply_to(m, nfe=20, wrt=m.fs.time, scheme="BACKWARD")

        m.fs.controller.gain_p.fix(-1e-1)
        m.fs.controller.gain_i.fix(-1e-2)
        m.fs.controller.setpoint.fix(5)
        m.fs.controller.mv_ref.fix(0.5)
        # m.fs.controller.integral_of_error[0].fix(0)

    m.fs.pump_to_tank = Arc(source=m.fs.pump.outlet, destination=m.fs.tank.inlet)
    m.fs.tank_to_valve = Arc(source=m.fs.tank.outlet, destination=m.fs.valve.inlet)
    pyo.TransformationFactory("network.expand_arcs").apply_to(m.fs)

    m.fs.pump.efficiency_isentropic.fix(0.8)
    m.fs.pump.deltaP.fix(5e4)

    m.fs.tank.tank_cross_sect_area.fix(20)
    m.fs.tank.tank_level[:].fix(5)

    m.fs.valve.Cv.fix(63.4676)
    m.fs.valve.valve_opening.fix(0.5)
    m.fs.pump.inlet.flow_mol[:].fix(7000)
    m.fs.pump.inlet.enth_mol[:].fix(3924)
    m.fs.pump.inlet.pressure[:].fix(1e7)

    m.fs.valve.outlet.pressure[:].fix(10050000)

    set_scaling_factors(m)

    if dynamic:
        m.fs.tank.set_initial_condition()
        m.fs.tank.tank_level.unfix()
        m.fs.tank.tank_level[0].fix()
        m.fs.valve.valve_opening.unfix()
    else:
        m.fs.pump.initialize()

        m.fs.tank.inlet.flow_mol[:].value = m.fs.pump.outlet.flow_mol[0].value
        m.fs.tank.inlet.enth_mol[:].value = m.fs.pump.outlet.enth_mol[0].value
        m.fs.tank.inlet.pressure[:].value = m.fs.pump.outlet.pressure[0].value

        m.fs.tank.initialize()

        m.fs.valve.inlet.flow_mol[:].value = m.fs.tank.outlet.flow_mol[0].value
        m.fs.valve.inlet.enth_mol[:].value = m.fs.tank.outlet.enth_mol[0].value
        m.fs.valve.inlet.pressure[:].value = m.fs.tank.outlet.pressure[0].value

        # Solve for outlet pressure
        m.fs.valve.initialize()

        m.fs.valve.valve_opening.unfix()
        dof = degrees_of_freedom(m)
        assert dof == 0
        # TODO: MA27 works better than MA57
        solver = get_solver(solver="ipopt_v2", solver_options={"linear_solver": "ma27"})
        solver.solve(m, tee=True)

    return m


def run_dynamic(m):
    # TODO: MA27 works better than MA57
    solver = get_solver(solver="ipopt_v2", solver_options={"linear_solver": "ma27"})

    # add step change
    for t in m.fs.time:
        if t >= 50:
            m.fs.pump.inlet.flow_mol[t].fix(m.fs.pump.inlet.flow_mol[0].value * 1.2)
        else:
            m.fs.pump.inlet.flow_mol[t].fix(m.fs.pump.inlet.flow_mol[0].value)
    solver.solve(m, tee=True)

    return m


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_pid(m):
    assert 0.5000 == pytest.approx(pyo.value(m.fs.valve.valve_opening[0.0]), abs=1e-3)
    assert 0.6156 == pytest.approx(
        pyo.value(m.fs.valve.valve_opening[406.25]), abs=1e-3
    )
    assert 0.60268 == pytest.approx(pyo.value(m.fs.valve.valve_opening[1000]), abs=1e-3)
    assert 5 == pytest.approx(pyo.value(m.fs.tank.tank_level[0.0]), abs=1e-3)
    assert 4.9762 == pytest.approx(pyo.value(m.fs.tank.tank_level[406.25]), abs=1e-3)
    assert 4.99923 == pytest.approx(pyo.value(m.fs.tank.tank_level[1000]), abs=1e-3)
    assert 7000.0 == pytest.approx(pyo.value(m.fs.tank.outlet.flow_mol[0.0]), abs=1e-3)
    assert 8598.66642 == pytest.approx(
        pyo.value(m.fs.tank.outlet.flow_mol[406.25]), abs=1e-3
    )
    assert 8436.87597 == pytest.approx(
        pyo.value(m.fs.tank.outlet.flow_mol[1000]), abs=1e-3
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Wrapper for Pyomo solvers to allow us to define default solver options
"""

from copy import deepcopy

from pyomo.environ import SolverFactory

import idaes


class SolverWrapper(object):
    """
    Wrapper for Pyomo solvers to allow us to define default solver options
    """

    def __init__(self, name, register=True):
        if name is None:
            name = "default"
        self.name = name
        self.registered = register
        if name == "default":
            self.solver = None
            doc = "IDAES Configured Default Solver"
        else:
            self.solver = SolverFactory.get_class(name)
            doc = SolverFactory.doc(name)
        if register:
            SolverFactory.unregister(name)
            # Re-register the solver (register is a decorator)
            SolverFactory.register(name, doc)(self)

    def __call__(self, *args, **kwargs):
        if self.name == "default":
            name = idaes.cfg.default_solver
            solver = SolverFactory.get_class(name)
        else:
            name = self.name
            solver = self.solver
        if name in idaes.cfg and (
            idaes.cfg.use_idaes_solver_config or not self.registered
        ):
            for k, v in idaes.cfg[name].items():
                if k not in kwargs:
                    kwargs[k] = deepcopy(v)
                elif k == "options":
                    # options is in ConfigBlock and in kwargs, treat "options"
                    # special so individual options can have defaults not just
                    # the whole options block
                    for opk, opv in v.items():
                        if opk not in kwargs["options"]:
                            kwargs["options"][opk] = opv
                elif k == "writer_config":
                    # writer_config is in ConfigBlock and in kwargs, treat "writer_config"
                    # special so individual options can have defaults not just
                    # the whole options block
                    for opk, opv in v.items():
                        if opk not in kwargs["writer_config"]:
                            kwargs["writer_config"][opk] = opv

        return solver(*args, **kwargs)


def use_idaes_solver_configuration_defaults(b=True):
    """
    This function enables (or disables if given False as the argument) solvers
    getting default settings from the IDAES configuration.  When enabled this
    allows global configuration of solvers.

    Args:
        b: True to use default solver configurations from the IDAES configuration
           False to use standard Pyomo solver factories. Default is True.

    Returns:
        None
    """
    idaes.cfg.use_idaes_solver_config = b
    if b:  # This will let you explicitly state you don't want any part of this
        # so if you only do "use_idaes_solver_configuration_defaults(False)" up-
        # front you are saying I know this stuff exists and I must insist you
        # don't use it, of course you can still implicitly not use it.  You can
        # also turn it off and on, if that makes sense for you, but once you turn
        # it on, you've still registered the wrapper classes, and if you turn
        # it off they just pass-through.
        for c in list(SolverFactory):
            if isinstance(SolverFactory.get_class(c), SolverWrapper):
                continue
            SolverWrapper(c)
        if "default" not in SolverFactory:
            SolverWrapper("default")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import pytest
import pandas as pd
import numpy as np
from pyomo.common import unittest as pyo_unittest
from idaes.apps.grid_integration.forecaster import PerfectForecaster
import idaes.logger as idaeslog


@pytest.fixture
def wind_df():
    """
    This is to define a dataframe fed to the PerfectForecaster with example DA/RT
    capacity factors and LMPs.
    """
    start_year = 2020
    start_mon = 1
    start_day = 1
    start_date = pd.Timestamp(f"{start_year}-{start_mon:02d}-{start_day:02d} 00:00:00")
    ix = pd.date_range(
        start=start_date,
        end=start_date + pd.offsets.DateOffset(days=1) - pd.offsets.DateOffset(hours=1),
        freq="1H",
    )
    df = pd.DataFrame(index=ix)
    df["303_WIND_1-DACF"] = list(i / 100 for i in range(24))
    df["303_WIND_1-RTCF"] = list((i + 1) / 100 for i in range(24))
    df["Caesar-DALMP"] = list(i for i in range(24))
    df["Caesar-RTLMP"] = list(i + 1 for i in range(24))
    return df


@pytest.fixture
def base_perfectforecaster(wind_df):
    return PerfectForecaster(wind_df)


@pytest.mark.component
def test_create_perfectforecaster(wind_df):
    perfectforecaster = PerfectForecaster(data_path_or_df=wind_df)
    assert perfectforecaster.data is wind_df


@pytest.mark.unit
@pytest.mark.parametrize("value", [np.array([1, 2, 3, 4, 5]), [1, 2, 3, 4, 5]])
def test_create_perfectforecaster_with_ndarray_and_list(value):
    """
    This is to test when we creat the PerfectForecaster with array or list, there will be an error.
    """
    with pytest.raises(
        ValueError,
        match=r".*The data_path_or_df should be pandas DataFrame or a string of the csv path.*",
    ):
        perfectforecaster = PerfectForecaster(value)


@pytest.mark.unit
def test_get_column_from_data(base_perfectforecaster):
    """
    This is to test the if the dataframe can be read correctly.
    """
    date = "2020-01-01"
    hour = 0
    horizon = 24
    col = "303_WIND_1-DACF"
    expected_forecast = [i / 100 for i in range(24)]
    result_forecast = base_perfectforecaster.get_column_from_data(
        date, hour, horizon, col
    )

    pyo_unittest.assertStructuredAlmostEqual(
        first=result_forecast.tolist(), second=expected_forecast
    )


@pytest.mark.unit
def test_forecast_day_ahead_prices(base_perfectforecaster):
    """
    This is to test if the forecast_day_ahead_price function works correctly.
    """
    date = "2020-01-01"
    hour = 0
    horizon = 24
    bus = "Caesar"
    nsp = 0  # this nsp is not used in the self.forecast_day_ahead_prices(), but we have it to make this func consistent with the stochastic bidder and coordinator.
    result_forecast = base_perfectforecaster.forecast_day_ahead_prices(
        date, hour, bus, horizon, nsp
    )
    expected_forecast = [i for i in range(24)]
    pyo_unittest.assertStructuredAlmostEqual(
        first=result_forecast.tolist(), second=expected_forecast
    )


@pytest.mark.unit
def test_forecast_real_time_prices(base_perfectforecaster):
    """
    This is to test if the forecast_real_time_price function works correctly.
    """
    date = "2020-01-01"
    hour = 0
    horizon = 4
    bus = "Caesar"
    nsp = 0
    result_forecast = base_perfectforecaster.forecast_real_time_prices(
        date, hour, bus, horizon, nsp
    )
    expected_forecast = [i + 1 for i in range(4)]
    pyo_unittest.assertStructuredAlmostEqual(
        first=result_forecast.tolist(), second=expected_forecast
    )


@pytest.mark.unit
def test_forecast_day_ahead_capacity_factor(base_perfectforecaster):
    """
    This is to test if the forecast_day_ahead_capacity_factor function works correctly.
    """
    date = "2020-01-01"
    hour = 0
    horizon = 24
    gen = "303_WIND_1"
    result_forecast = base_perfectforecaster.forecast_day_ahead_capacity_factor(
        date, hour, gen, horizon
    )
    expected_forecast = [i / 100 for i in range(24)]
    pyo_unittest.assertStructuredAlmostEqual(
        first=result_forecast.tolist(), second=expected_forecast
    )


@pytest.mark.unit
def test_forecast_real_time_capacity_factor(base_perfectforecaster):
    """
    This is to test if the forecast_real_time_capacity_factor function works correctly.
    """
    date = "2020-01-01"
    hour = 0
    horizon = 4
    gen = "303_WIND_1"
    result_forecast = base_perfectforecaster.forecast_real_time_capacity_factor(
        date, hour, gen, horizon
    )
    expected_forecast = [(i + 1) / 100 for i in range(4)]
    pyo_unittest.assertStructuredAlmostEqual(
        first=result_forecast.tolist(), second=expected_forecast
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
A unit model for a "node" that connects several pipelines,
possibly with its own supply and demand streams.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from pyomo.common.config import ConfigValue
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.var import Var
from pyomo.core.base.set import Set
from pyomo.core.base.block import Block
from pyomo.core.base.units_container import units as pyunits
from pyomo.core.base.reference import Reference
from pyomo.network.port import Port
from pyomo.network.arc import Arc

from idaes.core import declare_process_block_class, StateBlock, UnitModelBlockData
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.config import (
    is_physical_parameter_block,
)


@declare_process_block_class("PipelineNode")
class PipelineNodeData(UnitModelBlockData):

    CONFIG = UnitModelBlockData.CONFIG()

    CONFIG.declare(
        "property_package",
        ConfigValue(default=None, domain=is_physical_parameter_block),
    )
    CONFIG.declare(
        "n_inlet_pipelines",
        ConfigValue(default=1, domain=int),
    )
    CONFIG.declare(
        "n_outlet_pipelines",
        ConfigValue(default=1, domain=int),
    )
    CONFIG.declare(
        "n_supplies",
        ConfigValue(default=0, domain=int),
    )
    CONFIG.declare(
        "n_demands",
        ConfigValue(default=0, domain=int),
    )

    def build(self):
        super(PipelineNodeData, self).build()

        # self.config is the ConfigBlock "instantiated" from self.CONFIG
        # in ProcessBlockData.
        config = self.config
        time = self.flowsheet().time
        property_package = config.property_package

        if len(property_package.phase_list) != 1:
            raise ValueError(
                "%s can only be constructed with a "
                "single-phase property package.\n"
                "Got phases %s."
                % (self.__class__, [p for p in property_package.phase_list])
            )
        self.phase = next(iter(property_package.phase_list))
        if self.phase != "Vap":
            raise ValueError(
                '%s can only be constructed with a single phase, "Vap".'
                "Got phase %s." % (self.__class__, self.phase)
            )

        #
        # Make sure property package has the properties that this unit
        # model requires.
        #
        property_dict = property_package.get_metadata().properties
        if not property_dict["pressure"].supported:
            raise ValueError(
                "Property package supplied to pipeline must have a property "
                "for 'pressure', which was not found in %s." % type(property_package)
            )

        # Build a state block for this node. Note that the flow_mol attribute
        # of this state block should never be used. Should we delete it?
        # We should probably add a constraint that makes sure it is equal
        # to flow "through" this node.
        state_config = {"defined_state": True}
        self.state = property_package.build_state_block(time, **state_config)

        self.add_inlets()
        self.add_outlets()
        self.add_supplies()
        self.add_demands()
        self.add_flow_balance_con()
        self.add_total_flow_con()
        self.add_component_mixing_con()
        self.add_enthalpy_mixing_con()

        self.n_inlet_pipelines = 0
        self.n_outlet_pipelines = 0
        self._inlet_pipelines = {}
        self._outlet_pipelines = {}

    def inlet_pipelines(self):
        return self._inlet_pipelines

    def outlet_pipelines(self):
        return self._outlet_pipelines

    def get_inlet_pipeline(self, i):
        return self._inlet_pipelines[i]

    def get_outlet_pipeline(self, i):
        return self._outlet_pipelines[i]

    def get_port_name(self):
        return "port"

    def _get_port_and_references(self, name, block, doc=None):
        """
        This is a copy of UnitModel.add_port, except it returns components
        instead of adding them to self.
        """
        # Validate block object
        if not isinstance(block, StateBlock):
            raise ConfigurationError(
                "{} block object provided to add_port "
                "method is not an instance of a "
                "StateBlock object. IDAES port objects "
                "should only be associated with "
                "StateBlocks.".format(self.name)
            )

        # Create empty Port
        p = Port(doc=doc)
        # setattr(blk, name, p)

        # Get dict of Port members and names
        member_list = block[self.flowsheet().time.first()].define_port_members()

        # Create References for port members
        references = []
        for s in member_list:
            if not member_list[s].is_indexed():
                slicer = block[:].component(member_list[s].local_name)
            else:
                slicer = block[:].component(member_list[s].local_name)[...]

            r = Reference(slicer)
            # setattr(blk, "_"+s+"_"+name+"_ref", r)
            ref_name = "_" + s + "_" + name + "_ref"
            references.append((ref_name, r))

            # Add Reference to Port
            # Should this be deferred as well?
            p.add(r, s)

        return p, references

    def get_temperature_eq_con(self, state1, state2):
        time = self.flowsheet().time

        def temperature_eq_rule(b, t):
            return state1[t].temperature == state2[t].temperature

        return Constraint(time, rule=temperature_eq_rule)

    def get_pressure_eq_con(self, state1, state2):
        time = self.flowsheet().time

        def pressure_eq_rule(b, t):
            return state1[t].pressure == state2[t].pressure

        return Constraint(time, rule=pressure_eq_rule)

    def get_mole_frac_comp_eq_con(self, state1, state2):
        time = self.flowsheet().time
        comp_list = self.config.property_package.component_list

        def mole_frac_comp_eq_rule(b, t, j):
            return state1[t].mole_frac_comp[j] == state2[t].mole_frac_comp[j]

        return Constraint(time, comp_list, rule=mole_frac_comp_eq_rule)

    def get_port_block_rule(self, outlet=False):
        properties = self.config.property_package
        time = self.flowsheet().time
        state_config = {"defined_state": True}
        port_name = self.get_port_name()

        def block_rule(b, i):
            # Each inlet/outlet needs its own state block so it can have its
            # own flow rate, at the very least.
            b.state = properties.build_state_block(time, **state_config)
            port, refs = self._get_port_and_references(port_name, b.state)
            for ref_name, ref in refs:
                b.add_component(ref_name, ref)
            b.port = port
            b.has_pipeline = False

            # Add equality constraints between the node states and inlet/outlet
            # states for pressure and temperature.
            b.pressure_eq = self.get_pressure_eq_con(self.state, b.state)

            if outlet:
                # For inlets, mole fractions and temperatures are already
                # defined by mixing rules.
                b.mole_frac_comp_eq = self.get_mole_frac_comp_eq_con(
                    self.state, b.state
                )
                b.temperature_eq = self.get_temperature_eq_con(self.state, b.state)

        return block_rule

    def add_inlets(self):
        n_inlets = self.config.n_inlet_pipelines
        # TODO: Should these set indices be something other than integers?
        # For now, the same indices could be valid in inlets, outlets,
        # supplies, and demands, which could lead to hard-to-debug mistakes.
        self.inlet_set = Set(initialize=list(range(n_inlets)), dimen=1)
        rule = self.get_port_block_rule()
        self.inlets = Block(self.inlet_set, rule=rule)

    def add_outlets(self):
        n_outlets = self.config.n_outlet_pipelines
        self.outlet_set = Set(initialize=list(range(n_outlets)), dimen=1)
        rule = self.get_port_block_rule(outlet=True)
        self.outlets = Block(self.outlet_set, rule=rule)

    def get_supply_block_rule(self):
        """
        Add a block for each supply, which contains a state block,
        allowing us to specify different conditions for each supply.

        """
        time = self.flowsheet().time
        properties = self.config.property_package
        state_config = {"defined_state": True}

        def block_rule(b, i):
            # NOTE: Adding a state block here leaves us with temperature
            # and pressure variables that are unused. Strictly, pressure
            # should be set equal to the node's pressure, and temperature
            # should be used to calculate the node's temperature via an
            # energy balance.
            # We don't link pressure to avoid adding extra variables
            # (the node pressure may be used instead) and we don't have an
            # energy balance, so we leave temperature unconnected.
            b.state = properties.build_state_block(time, **state_config)
            # Add a reference here so we can access flow in the same way
            # as on demand blocks.
            b.flow_mol = Reference(b.state[:].flow_mol)

            # Add equation to link pressure of this supply
            # to that of the node.
            b.isobaric_eq = self.get_pressure_eq_con(self.state, b.state)

        return block_rule

    def get_demand_block_rule(self):
        """
        Add a block for each demand, which only contains a variable
        for flow rate. Other (intensive) variables have the same value
        as this node's state block.

        """
        # Should these demand blocks have references to the node's intensive
        # state variables? This would probably be convenient.
        time = self.flowsheet().time

        def block_rule(b, i):
            b.flow_mol = Var(time, initialize=100.0, units=pyunits.kmol / pyunits.hr)

        return block_rule

    def add_supplies(self):
        """
        Add set and indexed block for supplies.
        """
        n_supplies = self.config.n_supplies
        self.supply_set = Set(initialize=list(range(n_supplies)), dimen=1)
        rule = self.get_supply_block_rule()
        self.supplies = Block(self.supply_set, rule=rule)

    def add_demands(self):
        """
        Add set and indexed block for demands.
        """
        n_demands = self.config.n_demands
        self.demand_set = Set(initialize=list(range(n_demands)), dimen=1)
        # Should I give demands a state block? This seems redundant.
        rule = self.get_demand_block_rule()
        self.demands = Block(self.demand_set, rule=rule)

    def add_flow_balance_con(self):
        """
        Adds a total flow rate balance stating that inlet flow must equal
        outlet flow.

        """
        # Should this be a balance on component flow rates instead?
        time = self.flowsheet().time

        def flow_balance_rule(b, t):
            return sum(self.supplies[:].flow_mol[t]) + sum(
                self.inlets[:].state[t].flow_mol
            ) == sum(self.demands[:].flow_mol[t]) + sum(
                self.outlets[:].state[t].flow_mol
            )

        self.flow_balance = Constraint(time, rule=flow_balance_rule)

    def add_total_flow_con(self):
        """
        Calculates the flow rate through this node as the sum of flow rates
        from supplies and inlet pipelines.

        """
        time = self.flowsheet().time

        def total_flow_rule(b, t):
            return self.state[t].flow_mol == sum(self.supplies[:].flow_mol[t]) + sum(
                self.inlets[:].state[t].flow_mol
            )

        self.total_flow_eq = Constraint(time, rule=total_flow_rule)

    def add_component_mixing_con(self):
        """
        Adds a component mixing equation that calculates mole fractions
        of the node from mole fractions and flow rates of the supplies
        and inlet pipelines.

        """
        time = self.flowsheet().time
        component_list = self.config.property_package.component_list

        def component_mixing_rule(b, t, j):
            return (
                sum(self.supplies[i].state[t].flow_mol_comp[j] for i in self.supply_set)
                + sum(self.inlets[i].state[t].flow_mol_comp[j] for i in self.inlet_set)
                == self.state[t].flow_mol_comp[j]
            )

        self.component_mixing_eq = Constraint(
            time, component_list, rule=component_mixing_rule
        )

    def add_enthalpy_mixing_con(self):
        """
        Adds an equation that calculates temperature of the node from
        temperatures, heat capacities, and flow rates of the supplies
        and inlet pipelines.

        """
        time = self.flowsheet().time
        # We enforce that phase_list contains only "Vap"
        p = next(iter(self.config.property_package.phase_list))

        def enthalpy_mixing_rule(b, t):
            supplies = [self.supplies[i].state[t] for i in self.supply_set]
            inlets = [self.inlets[i].state[t] for i in self.inlet_set]
            return sum(supply.get_enthalpy_flow_terms(p) for supply in supplies) + sum(
                inlet.get_enthalpy_flow_terms(p) for inlet in inlets
            ) == self.state[t].get_enthalpy_flow_terms(p)

        self.enthalpy_mixing_eq = Constraint(time, rule=enthalpy_mixing_rule)

    def add_pipeline_to_inlet(self, pipeline, idx=None):
        """
        Creates an Arc between the outlet port of a pipeline and one
        of this node's inlet ports.

        """
        pipeline_port = pipeline.outlet_port
        if idx is None:
            idx = self.n_inlet_pipelines
        if idx not in self.inlet_set:
            # Either bad index given or no more inlets
            raise ValueError()
        if self.inlets[idx].has_pipeline:
            # Don't want to add two pipelines to the same inlet
            raise RuntimeError()
        self.inlets[idx].arc = Arc(ports=(pipeline_port, self.inlets[idx].port))
        self.inlets[idx].has_pipeline = True
        self._inlet_pipelines[idx] = pipeline
        self.n_inlet_pipelines += 1

    def add_pipeline_to_outlet(self, pipeline, idx=None):
        """
        Adds an Arc between one of this node's outlet ports
        and the inlet port of a pipeline.

        """
        pipeline_port = pipeline.inlet_port
        if idx is None:
            idx = self.n_outlet_pipelines
        if idx not in self.outlet_set:
            # Either bad index given or no more outlets
            raise ValueError()
        if self.outlets[idx].has_pipeline:
            # Don't want to add two pipelines to the same outlet
            raise RuntimeError()
        self.outlets[idx].arc = Arc(ports=(pipeline_port, self.outlets[idx].port))
        self.outlets[idx].has_pipeline = True
        self._outlet_pipelines[idx] = pipeline
        self.n_outlet_pipelines += 1



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test for ControlVolumeBlockData, and for initializing the
bubbling fluidized bed module

Author: Chinedu Okoli
"""

import pytest

from pyomo.environ import (
    ConcreteModel,
    check_optimal_termination,
    SolverStatus,
    value,
    Var,
    Constraint,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.config import ConfigBlock
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
    unused_variables_set,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver
from idaes.core.util import scaling as iscale
from idaes.core.util.exceptions import InitializationError

from idaes.models_extra.gas_solid_contactors.unit_models.bubbling_fluidized_bed import (
    BubblingFluidizedBed,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.gas_phase_thermo import (
    GasPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.solid_phase_thermo import (
    SolidPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.hetero_reactions import (
    HeteroReactionParameterBlock,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # Set up thermo props and reaction props
    m.fs.gas_properties = GasPhaseParameterBlock()
    m.fs.solid_properties = SolidPhaseParameterBlock()
    m.fs.hetero_reactions = HeteroReactionParameterBlock(
        solid_property_package=m.fs.solid_properties,
        gas_property_package=m.fs.gas_properties,
    )

    m.fs.unit = BubblingFluidizedBed(
        gas_phase_config={"property_package": m.fs.gas_properties},
        solid_phase_config={
            "property_package": m.fs.solid_properties,
            "reaction_package": m.fs.hetero_reactions,
        },
    )

    # Check unit config arguments
    assert len(m.fs.unit.config) == 14
    assert isinstance(m.fs.unit.config.gas_phase_config, ConfigBlock)
    assert isinstance(m.fs.unit.config.solid_phase_config, ConfigBlock)

    assert m.fs.unit.config.finite_elements == 5
    assert m.fs.unit.config.length_domain_set == [0.0, 1.0]
    assert m.fs.unit.config.transformation_method == "dae.finite_difference"
    assert m.fs.unit.config.transformation_scheme == "BACKWARD"
    assert m.fs.unit.config.collocation_points == 3
    assert m.fs.unit.config.flow_type == "co_current"
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.componentTotal
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.enthalpyTotal
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.has_pressure_change is True

    # Check gas phase config arguments
    assert len(m.fs.unit.config.gas_phase_config) == 7
    assert m.fs.unit.config.gas_phase_config.has_equilibrium_reactions is False
    assert m.fs.unit.config.gas_phase_config.property_package is m.fs.gas_properties
    assert m.fs.unit.config.gas_phase_config.reaction_package is None

    # Check solid phase config arguments
    assert len(m.fs.unit.config.solid_phase_config) == 7
    assert m.fs.unit.config.solid_phase_config.has_equilibrium_reactions is False
    assert m.fs.unit.config.solid_phase_config.property_package is m.fs.solid_properties
    assert m.fs.unit.config.solid_phase_config.reaction_package is m.fs.hetero_reactions


# -----------------------------------------------------------------------------
class TestIronOC(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = BubblingFluidizedBed(
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # Fix geometry variables
        m.fs.unit.number_orifice.fix(2500)  # [-]
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(272.81)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(373)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(1.86e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.4772)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.0646)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.4582)

        m.fs.unit.solid_inlet.flow_mass[0].fix(1422)  # kg/s
        m.fs.unit.solid_inlet.particle_porosity[0].fix(0.27)  # (-)
        m.fs.unit.solid_inlet.temperature[0].fix(1186)  # K
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iron_oc):
        assert hasattr(iron_oc.fs.unit, "gas_inlet")
        assert len(iron_oc.fs.unit.gas_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_inlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_inlet")
        assert len(iron_oc.fs.unit.solid_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_inlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.temperature, Var)

        assert hasattr(iron_oc.fs.unit, "gas_outlet")
        assert len(iron_oc.fs.unit.gas_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_outlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_outlet")
        assert len(iron_oc.fs.unit.solid_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_outlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.temperature, Var)

        assert isinstance(iron_oc.fs.unit.orifice_area, Constraint)
        assert isinstance(iron_oc.fs.unit.bed_area_eqn, Constraint)
        assert isinstance(iron_oc.fs.unit.bubble_area, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_emulsion_area, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_emulsion_area, Constraint)
        assert isinstance(iron_oc.fs.unit.bubble_cloud_heat_trans_coeff, Constraint)
        assert isinstance(iron_oc.fs.unit.convective_heat_trans_coeff, Constraint)
        assert isinstance(iron_oc.fs.unit.convective_heat_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.bubble_cloud_bulk_heat_trans, Constraint)
        assert isinstance(iron_oc.fs.unit.bubble_mass_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_emulsion_mass_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.bubble_heat_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_emulsion_heat_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_emulsion_heat_transfer, Constraint)

        assert number_variables(iron_oc) == 809
        assert number_total_constraints(iron_oc) == 764
        assert number_unused_variables(iron_oc) == 19

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.fixture(scope="class")
    def iron_oc_unscaled(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = BubblingFluidizedBed(
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # Fix geometry variables
        m.fs.unit.number_orifice.fix(2500)  # [-]
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(272.81)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(373)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(1.86e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.4772)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.0646)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.4582)

        m.fs.unit.solid_inlet.flow_mass[0].fix(1422)  # kg/s
        m.fs.unit.solid_inlet.particle_porosity[0].fix(0.27)  # (-)
        m.fs.unit.solid_inlet.temperature[0].fix(1186)  # K
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_unscaled(self, iron_oc_unscaled):
        initialization_tester(
            iron_oc_unscaled,
            optarg={"tol": 1e-6},
            gas_phase_state_args={
                "flow_mol": 272.81,
                "temperature": 1186,
                "pressure": 1.86e5,
            },
            solid_phase_state_args={"flow_mass": 1230, "temperature": 1186},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_unscaled(self, iron_oc_unscaled):
        results = solver.solve(iron_oc_unscaled)

        # Check for optimal solution
        assert check_optimal_termination(results)
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_scaling(self, iron_oc):
        BFB = iron_oc.fs.unit  # alias to keep test lines short
        iscale.calculate_scaling_factors(BFB)

        assert pytest.approx(0.01538, abs=1e-5) == iscale.get_scaling_factor(
            BFB.bed_diameter
        )
        assert pytest.approx(0.00301, abs=1e-5) == iscale.get_scaling_factor(
            BFB.bed_area
        )
        for (t, x), v in BFB.delta.items():
            assert pytest.approx(1.00000, abs=1e-5) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB.bubble_diameter_max.items():
            assert pytest.approx(0.15385, abs=1e-5) == iscale.get_scaling_factor(v)
        for (t, x, j), v in BFB.Kgbulk_c.items():
            assert pytest.approx(3.014e-5, abs=1e-7) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB._reform_var_1.items():
            assert pytest.approx(65.00000, abs=1e-5) == iscale.get_scaling_factor(v)
        for (t, x, j), v in BFB._reform_var_2.items():
            assert pytest.approx(16927.95406, abs=1e-5) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB._reform_var_3.items():
            assert pytest.approx(2.83941, abs=1e-5) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB.Hbe.items():
            assert pytest.approx(1.000e-9, abs=1e-11) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB.Hgbulk.items():
            assert pytest.approx(1.000e-9, abs=1e-8) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB._reform_var_4.items():
            assert pytest.approx(1.093e-9, abs=1e-8) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB._reform_var_5.items():
            assert pytest.approx(6.667e-6, abs=1e-8) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB.htc_conv.items():
            assert pytest.approx(6.667e-6, abs=1e-5) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB.ht_conv.items():
            assert pytest.approx(6.667e-6, abs=1e-5) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB.bubble.heat.items():
            assert pytest.approx(1e-9, abs=1e-11) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB.gas_emulsion.heat.items():
            assert pytest.approx(1e-9, abs=1e-11) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB.solid_emulsion.heat.items():
            assert pytest.approx(1e-9, abs=1e-11) == iscale.get_scaling_factor(v)

        for c in BFB.orifice_area.values():
            assert pytest.approx(
                0.00301, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for c in BFB.bed_area_eqn.values():
            assert pytest.approx(
                0.00301, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_area.items():
            assert pytest.approx(
                0.00301, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.gas_emulsion_area.items():
            assert pytest.approx(
                0.00301, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.solid_emulsion_area.items():
            assert pytest.approx(
                0.00301, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.emulsion_vol_frac.items():
            assert pytest.approx(
                1.0000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.average_voidage.items():
            assert pytest.approx(
                1.0000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_growth_coefficient.items():
            assert pytest.approx(
                0.01538, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_diameter_maximum.items():
            assert pytest.approx(
                8.619e-10, abs=1e-12
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB._reformulation_eqn_1.items():
            assert pytest.approx(
                6.50000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_diameter_eqn.items():
            assert pytest.approx(
                0.01538, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.ddia_bubbledx_disc_eq.items():
            assert pytest.approx(
                0.01538, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_velocity_rise.items():
            assert pytest.approx(
                0.15385, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.emulsion_voidage.items():
            assert pytest.approx(
                1.00000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_velocity.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.average_gas_density_eqn.items():
            assert pytest.approx(
                1e-6, abs=1e-8
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.velocity_gas_superficial.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_vol_frac_eqn.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.solid_super_vel.items():
            assert pytest.approx(
                0.00001, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.gas_emulsion_pressure_drop.items():
            assert pytest.approx(
                1e-5, abs=1e-7
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB._reformulation_eqn_2.items():
            assert pytest.approx(
                169.27954, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB._reformulation_eqn_3.items():
            assert pytest.approx(
                2.83941, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB.bubble_cloud_mass_trans_coeff.items():
            assert pytest.approx(
                2.83941, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB.bubble_cloud_bulk_mass_trans.items():
            assert pytest.approx(
                3.014e-5, abs=1e-7
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB.bubble_mass_transfer.items():
            assert pytest.approx(
                3.014e-5, abs=1e-7
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB.gas_emulsion_mass_transfer.items():
            assert pytest.approx(
                3.014e-5, abs=1e-7
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, r), c in BFB.solid_emulsion_rxn_ext_constraint.items():
            assert pytest.approx(
                3.01358e-5, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB.gas_emulsion_hetero_rxn_eqn.items():
            assert pytest.approx(
                3.01358e-5, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_gas_flowrate.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.emulsion_gas_flowrate.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.gas_emulsion_pressure_in.items():
            assert pytest.approx(
                1e-5, abs=1e-7
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.gas_mole_flow_in.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.particle_porosity_in.items():
            assert pytest.approx(
                100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.emulsion_gas_velocity_in.items():
            assert pytest.approx(
                25.23722, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, j), c in BFB.bubble_mole_frac_in.items():
            assert pytest.approx(
                10.00000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, j), c in BFB.gas_emulsion_mole_frac_in.items():
            assert pytest.approx(
                10.00000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.solid_emulsion_mass_flow_in.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, j), c in BFB.solid_emulsion_mass_frac_in.items():
            assert pytest.approx(
                10.00000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.gas_pressure_out.items():
            assert pytest.approx(
                1e-5, abs=1e-7
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, p, j), c in BFB.gas_material_balance_out.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, p, j), c in BFB.solid_material_balance_out.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.solid_particle_porosity_out.items():
            assert pytest.approx(
                100.00000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB._reformulation_eqn_4.items():
            assert pytest.approx(
                0, abs=1e-10
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB._reformulation_eqn_5.items():
            assert pytest.approx(
                0.66666, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_cloud_heat_trans_coeff.items():
            assert pytest.approx(
                1e-11, abs=1e-11
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.convective_heat_trans_coeff.items():
            assert pytest.approx(
                6.667e-6, abs=1e-8
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.convective_heat_transfer.items():
            assert pytest.approx(
                6.667e-8, abs=1e-10
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_cloud_bulk_heat_trans.items():
            assert pytest.approx(
                1e-9, abs=1e-11
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_heat_transfer.items():
            assert pytest.approx(
                1e-11, abs=1e-11
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.gas_emulsion_heat_transfer.items():
            if x == 0:
                assert pytest.approx(
                    1e-11, abs=1e-11
                ) == iscale.get_constraint_transform_applied_scaling_factor(c)
            else:
                assert pytest.approx(
                    1e-9, abs=1e-11
                ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.solid_emulsion_heat_transfer.items():
            if x == 0:
                assert pytest.approx(
                    1e-11, abs=1e-11
                ) == iscale.get_constraint_transform_applied_scaling_factor(c)
            else:
                assert pytest.approx(
                    1e-9, abs=1e-11
                ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, p), c in BFB.gas_energy_balance_in.items():
            assert pytest.approx(
                1e-9, abs=1e-11
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.gas_emulsion_temperature_in.items():
            assert pytest.approx(
                0.01000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.solid_energy_balance_in.items():
            assert pytest.approx(
                1e-9, abs=1e-11
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.gas_energy_balance_out.items():
            assert pytest.approx(
                1e-9, abs=1e-11
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.solid_energy_balance_out.items():
            assert pytest.approx(
                1e-9, abs=1e-11
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        BFB = iron_oc.fs.unit  # alias to keep test lines short
        iscale.calculate_scaling_factors(BFB)
        initialization_tester(
            iron_oc,
            optarg={"tol": 1e-6},
            gas_phase_state_args={
                "flow_mol": 272.81,
                "temperature": 1186,
                "pressure": 1.86e5,
            },
            solid_phase_state_args={"flow_mass": 1230, "temperature": 1186},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iron_oc):
        results = solver.solve(iron_oc, tee=True)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iron_oc):
        assert (
            pytest.approx(0.13964, abs=1e-5)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 0].value
        )
        assert (
            pytest.approx(1.02772, abs=1e-5)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 1].value
        )
        assert (
            pytest.approx(0.01522, abs=1e-5)
            == iron_oc.fs.unit.bubble_diameter[0, 0].value
        )
        assert (
            pytest.approx(1.02990, abs=1e-5)
            == iron_oc.fs.unit.bubble_diameter[0, 1].value
        )
        assert (
            pytest.approx(0.37470, abs=1e-5)
            == iron_oc.fs.unit.velocity_bubble[0, 0].value
        )
        assert (
            pytest.approx(3.24767, abs=1e-5)
            == iron_oc.fs.unit.velocity_bubble[0, 1].value
        )
        assert pytest.approx(0.266906, abs=1e-5) == iron_oc.fs.unit.delta[0, 0].value
        assert pytest.approx(0.30425, abs=1e-5) == iron_oc.fs.unit.delta[0, 1].value
        assert (
            pytest.approx(123088.236424, rel=1e-5)
            == iron_oc.fs.unit.gas_outlet.pressure[0].value
        )
        assert (
            pytest.approx(62911.76357, rel=1e-5)
            == iron_oc.fs.unit.gas_inlet.pressure[0].value
            - iron_oc.fs.unit.gas_outlet.pressure[0].value
        )

    @pytest.mark.component
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iron_oc):
        # Conservation of material check
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_inlet_block[0].mw,
            iron_oc.fs.unit.gas_inlet_block[0].mw_eqn,
        )
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_outlet_block[0].mw,
            iron_oc.fs.unit.gas_outlet_block[0].mw_eqn,
        )
        mbal_gas = value(
            (
                iron_oc.fs.unit.gas_inlet.flow_mol[0]
                * iron_oc.fs.unit.gas_inlet_block[0].mw
            )
            - (
                iron_oc.fs.unit.gas_outlet.flow_mol[0]
                * iron_oc.fs.unit.gas_outlet_block[0].mw
            )
        )
        mbal_solid = value(
            iron_oc.fs.unit.solid_inlet.flow_mass[0]
            - iron_oc.fs.unit.solid_outlet.flow_mass[0]
        )
        mbal_tol = mbal_gas + mbal_solid
        assert abs(mbal_tol) <= 1e-2

        # Reaction stoichiometric ratio check
        # Overall reducer reactions for methane combustion:
        # CH4 + 12Fe2O3 => 8Fe3O4 + CO2 + 2H2O
        mole_gas_reacted = value(
            iron_oc.fs.unit.gas_inlet.flow_mol[0]
            * iron_oc.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"]
            - iron_oc.fs.unit.gas_outlet.flow_mol[0]
            * iron_oc.fs.unit.gas_outlet.mole_frac_comp[0, "CH4"]
        )
        mole_solid_reacted = value(
            (
                iron_oc.fs.unit.solid_inlet.flow_mass[0]
                * iron_oc.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_inlet_block[0]._params.mw_comp["Fe2O3"]
            )
            - (
                iron_oc.fs.unit.solid_outlet.flow_mass[0]
                * iron_oc.fs.unit.solid_outlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_outlet_block[0]._params.mw_comp["Fe2O3"]
            )
        )
        stoichiometric_ratio = mole_solid_reacted / mole_gas_reacted
        assert pytest.approx(12, abs=1e-6) == stoichiometric_ratio

        # Conservation of energy check
        ebal_gas = value(
            (
                iron_oc.fs.unit.gas_inlet.flow_mol[0]
                * iron_oc.fs.unit.gas_inlet_block[0].enth_mol
            )
            - (
                iron_oc.fs.unit.gas_outlet.flow_mol[0]
                * iron_oc.fs.unit.gas_outlet_block[0].enth_mol
            )
        )
        ebal_solid = value(
            (
                iron_oc.fs.unit.solid_inlet.flow_mass[0]
                * iron_oc.fs.unit.solid_inlet_block[0].enth_mass
            )
            - (
                iron_oc.fs.unit.solid_outlet.flow_mass[0]
                * iron_oc.fs.unit.solid_outlet_block[0].enth_mass
            )
        )
        e_reaction = value(
            mole_gas_reacted
            * iron_oc.fs.unit.solid_emulsion.reactions[0, 0]._params.dh_rxn["R1"]
        )
        ebal_tol = ebal_gas + ebal_solid - e_reaction
        assert abs(ebal_tol) <= 1e-2

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        iron_oc.fs.unit.report()

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        iron_oc.fs.unit.gas_outlet.flow_mol[0].fix(1)

        with pytest.raises(InitializationError):
            iron_oc.fs.unit.initialize()


# -----------------------------------------------------------------------------
class TestIronOC_EnergyBalanceType(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = BubblingFluidizedBed(
            energy_balance_type=EnergyBalanceType.none,
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # # Fix geometry variables
        m.fs.unit.number_orifice.fix(2500)  # [-]
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(272.81)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(1186)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(1.86e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.4772)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.0646)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.4582)

        m.fs.unit.solid_inlet.flow_mass[0].fix(1422)  # kg/
        m.fs.unit.solid_inlet.particle_porosity.fix(0.27)  # (-)
        m.fs.unit.solid_inlet.temperature[0].fix(1186)  # K
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iron_oc):
        assert hasattr(iron_oc.fs.unit, "gas_inlet")
        assert len(iron_oc.fs.unit.gas_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_inlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_inlet")
        assert len(iron_oc.fs.unit.solid_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_inlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.temperature, Var)

        assert hasattr(iron_oc.fs.unit, "gas_outlet")
        assert len(iron_oc.fs.unit.gas_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_outlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_outlet")
        assert len(iron_oc.fs.unit.solid_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_outlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.temperature, Var)

        assert isinstance(iron_oc.fs.unit.gas_energy_balance_out, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_energy_balance_out, Constraint)
        assert isinstance(iron_oc.fs.unit.isothermal_gas_emulsion, Constraint)
        assert isinstance(iron_oc.fs.unit.isothermal_solid_emulsion, Constraint)
        assert isinstance(iron_oc.fs.unit.isothermal_bubble, Constraint)

        assert number_variables(iron_oc) == 649
        assert number_total_constraints(iron_oc) == 571
        assert number_unused_variables(iron_oc) == 53
        print(unused_variables_set(iron_oc))

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.fixture(scope="class")
    def iron_oc_unscaled(self, iron_oc):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = BubblingFluidizedBed(
            energy_balance_type=EnergyBalanceType.none,
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # # Fix geometry variables
        m.fs.unit.number_orifice.fix(2500)  # [-]
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(272.81)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(1186)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(1.86e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.4772)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.0646)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.4582)

        m.fs.unit.solid_inlet.flow_mass[0].fix(1422)  # kg/
        m.fs.unit.solid_inlet.particle_porosity.fix(0.27)  # (-)
        m.fs.unit.solid_inlet.temperature[0].fix(1186)  # K
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_unscaled(self, iron_oc_unscaled):
        initialization_tester(
            iron_oc_unscaled,
            optarg={"tol": 1e-6},
            gas_phase_state_args={
                "flow_mol": 272.81,
                "temperature": 1186,
                "pressure": 1.86e5,
            },
            solid_phase_state_args={"flow_mass": 1422, "temperature": 1186},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_unscaled(self, iron_oc_unscaled):
        results = solver.solve(iron_oc_unscaled)

        # Check for optimal solution
        assert check_optimal_termination(results)
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_scaling(self, iron_oc):
        BFB = iron_oc.fs.unit  # alias to keep test lines short
        iscale.calculate_scaling_factors(BFB)

        assert pytest.approx(0.01538, abs=1e-5) == iscale.get_scaling_factor(
            BFB.bed_diameter
        )
        assert pytest.approx(0.00301, abs=1e-5) == iscale.get_scaling_factor(
            BFB.bed_area
        )
        for (t, x), v in BFB.delta.items():
            assert pytest.approx(1.00000, abs=1e-5) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB.bubble_diameter_max.items():
            assert pytest.approx(0.15385, abs=1e-5) == iscale.get_scaling_factor(v)
        for (t, x, j), v in BFB.Kgbulk_c.items():
            assert pytest.approx(3.014e-5, abs=1e-5) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB._reform_var_1.items():
            assert pytest.approx(65.00000, rel=1e-5) == iscale.get_scaling_factor(v)
        for (t, x, j), v in BFB._reform_var_2.items():
            assert pytest.approx(16927.95406, rel=1e-5) == iscale.get_scaling_factor(v)
        for (t, x), v in BFB._reform_var_3.items():
            assert pytest.approx(2.83941, rel=1e-5) == iscale.get_scaling_factor(v)

        for c in BFB.orifice_area.values():
            assert pytest.approx(
                3.014e-3, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for c in BFB.bed_area_eqn.values():
            assert pytest.approx(
                3.014e-3, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_area.items():
            assert pytest.approx(
                3.014e-3, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.gas_emulsion_area.items():
            assert pytest.approx(
                3.014e-3, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.solid_emulsion_area.items():
            assert pytest.approx(
                3.014e-3, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.emulsion_vol_frac.items():
            assert pytest.approx(
                1.00000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.average_voidage.items():
            assert pytest.approx(
                1.00000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_growth_coefficient.items():
            assert pytest.approx(
                0.015385, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_diameter_maximum.items():
            assert pytest.approx(
                8.619e-10, abs=1e-12
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB._reformulation_eqn_1.items():
            assert pytest.approx(
                6.50000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_diameter_eqn.items():
            assert pytest.approx(
                0.01538, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.ddia_bubbledx_disc_eq.items():
            assert pytest.approx(
                0.01538, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_velocity_rise.items():
            assert pytest.approx(
                0.15385, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.emulsion_voidage.items():
            assert pytest.approx(
                1.00000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_velocity.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.average_gas_density_eqn.items():
            assert pytest.approx(
                1e-6, abs=1e-8
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.velocity_gas_superficial.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_vol_frac_eqn.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.solid_super_vel.items():
            assert pytest.approx(
                0.00001, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.gas_emulsion_pressure_drop.items():
            assert pytest.approx(
                1e-5, abs=1e-7
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB._reformulation_eqn_2.items():
            assert pytest.approx(
                169.27954, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB._reformulation_eqn_3.items():
            assert pytest.approx(
                2.83941, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB.bubble_cloud_mass_trans_coeff.items():
            assert pytest.approx(
                2.83941, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB.bubble_cloud_bulk_mass_trans.items():
            assert pytest.approx(
                3.014e-5, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB.bubble_mass_transfer.items():
            assert pytest.approx(
                3.014e-5, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB.gas_emulsion_mass_transfer.items():
            assert pytest.approx(
                3.014e-5, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, r), c in BFB.solid_emulsion_rxn_ext_constraint.items():
            assert pytest.approx(
                3.01358e-5, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in BFB.gas_emulsion_hetero_rxn_eqn.items():
            assert pytest.approx(
                3.01358e-5, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.bubble_gas_flowrate.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in BFB.emulsion_gas_flowrate.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.gas_emulsion_pressure_in.items():
            assert pytest.approx(
                1e-5, abs=1e-7
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.gas_mole_flow_in.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.particle_porosity_in.items():
            assert pytest.approx(
                100.00000, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.emulsion_gas_velocity_in.items():
            assert pytest.approx(
                25.23722, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, j), c in BFB.bubble_mole_frac_in.items():
            assert pytest.approx(
                10.00000, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, j), c in BFB.gas_emulsion_mole_frac_in.items():
            assert pytest.approx(
                10.00000, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.solid_emulsion_mass_flow_in.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, j), c in BFB.solid_emulsion_mass_frac_in.items():
            assert pytest.approx(
                10.00000, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.gas_pressure_out.items():
            assert pytest.approx(
                1e-5, abs=1e-7
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, p, j), c in BFB.gas_material_balance_out.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, p, j), c in BFB.solid_material_balance_out.items():
            assert pytest.approx(
                0.00100, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.solid_particle_porosity_out.items():
            assert pytest.approx(
                100.00000, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.gas_energy_balance_out.items():
            assert pytest.approx(
                0.01, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in BFB.solid_energy_balance_out.items():
            assert pytest.approx(
                0.01, abs=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        BFB = iron_oc.fs.unit  # alias to keep test lines short
        iscale.calculate_scaling_factors(BFB)
        initialization_tester(
            iron_oc,
            optarg={"tol": 1e-6},
            gas_phase_state_args={
                "flow_mol": 272.81,
                "temperature": 1186,
                "pressure": 1.86e5,
            },
            solid_phase_state_args={"flow_mass": 1422, "temperature": 1186},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iron_oc):
        results = solver.solve(iron_oc, tee=True)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iron_oc):
        assert (
            pytest.approx(0.44398, abs=1e-5)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 0].value
        )
        assert (
            pytest.approx(1.01868, abs=1e-5)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 1].value
        )
        assert (
            pytest.approx(0.02661, abs=1e-5)
            == iron_oc.fs.unit.bubble_diameter[0, 0].value
        )
        assert (
            pytest.approx(1.02861, abs=1e-5)
            == iron_oc.fs.unit.bubble_diameter[0, 1].value
        )
        assert (
            pytest.approx(0.76758, abs=1e-5)
            == iron_oc.fs.unit.velocity_bubble[0, 0].value
        )
        assert (
            pytest.approx(3.23723, abs=1e-5)
            == iron_oc.fs.unit.velocity_bubble[0, 1].value
        )
        assert pytest.approx(0.52679, abs=1e-5) == iron_oc.fs.unit.delta[0, 0].value
        assert pytest.approx(0.30244, abs=1e-5) == iron_oc.fs.unit.delta[0, 1].value
        assert (
            pytest.approx(122577.18737, rel=1e-4)
            == iron_oc.fs.unit.gas_outlet.pressure[0].value
        )
        assert (
            pytest.approx(63422.81262, rel=1e-4)
            == iron_oc.fs.unit.gas_inlet.pressure[0].value
            - iron_oc.fs.unit.gas_outlet.pressure[0].value
        )

    @pytest.mark.component
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iron_oc):
        # Conservation of material check
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_inlet_block[0].mw,
            iron_oc.fs.unit.gas_inlet_block[0].mw_eqn,
        )
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_outlet_block[0].mw,
            iron_oc.fs.unit.gas_outlet_block[0].mw_eqn,
        )
        mbal_gas = value(
            (
                iron_oc.fs.unit.gas_inlet.flow_mol[0]
                * iron_oc.fs.unit.gas_inlet_block[0].mw
            )
            - (
                iron_oc.fs.unit.gas_outlet.flow_mol[0]
                * iron_oc.fs.unit.gas_outlet_block[0].mw
            )
        )
        mbal_solid = value(
            iron_oc.fs.unit.solid_inlet.flow_mass[0]
            - iron_oc.fs.unit.solid_outlet.flow_mass[0]
        )
        mbal_tol = mbal_gas + mbal_solid
        assert abs(mbal_tol) <= 1e-2

        # Reaction stoichiometric ratio check
        # Overall reducer reactions for methane combustion:
        # CH4 + 12Fe2O3 => 8Fe3O4 + CO2 + 2H2O
        mole_gas_reacted = value(
            iron_oc.fs.unit.gas_inlet.flow_mol[0]
            * iron_oc.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"]
            - iron_oc.fs.unit.gas_outlet.flow_mol[0]
            * iron_oc.fs.unit.gas_outlet.mole_frac_comp[0, "CH4"]
        )
        mole_solid_reacted = value(
            (
                iron_oc.fs.unit.solid_inlet.flow_mass[0]
                * iron_oc.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_inlet_block[0]._params.mw_comp["Fe2O3"]
            )
            - (
                iron_oc.fs.unit.solid_outlet.flow_mass[0]
                * iron_oc.fs.unit.solid_outlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_outlet_block[0]._params.mw_comp["Fe2O3"]
            )
        )
        stoichiometric_ratio = mole_solid_reacted / mole_gas_reacted
        assert pytest.approx(12, abs=1e-6) == stoichiometric_ratio

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        iron_oc.fs.unit.report()

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        iron_oc.fs.unit.gas_outlet.flow_mol[0].fix(1)

        with pytest.raises(InitializationError):
            iron_oc.fs.unit.initialize()


# -----------------------------------------------------------------------------
class TestIronOC_TransformationMethod(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Set up thermo props and reaction props
        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.hetero_reactions = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )

        m.fs.unit = BubblingFluidizedBed(
            transformation_method="dae.collocation",
            gas_phase_config={"property_package": m.fs.gas_properties},
            solid_phase_config={
                "property_package": m.fs.solid_properties,
                "reaction_package": m.fs.hetero_reactions,
            },
        )

        # Fix geometry variables
        m.fs.unit.number_orifice.fix(2500)  # [-]
        m.fs.unit.bed_diameter.fix(6.5)  # m
        m.fs.unit.bed_height.fix(5)  # m

        # Fix inlet port variables for gas and solid
        m.fs.unit.gas_inlet.flow_mol[0].fix(272.81)  # mol/s
        m.fs.unit.gas_inlet.temperature[0].fix(373)  # K
        m.fs.unit.gas_inlet.pressure[0].fix(1.86e5)  # Pa = 1E5 bar
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.4772)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.0646)
        m.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.4582)

        m.fs.unit.solid_inlet.flow_mass[0].fix(1422)  # kg/s
        m.fs.unit.solid_inlet.particle_porosity[0].fix(0.27)  # (-)
        m.fs.unit.solid_inlet.temperature[0].fix(1186)  # K
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iron_oc):
        assert iron_oc.fs.unit.config.transformation_method == "dae.collocation"
        assert hasattr(iron_oc.fs.unit, "gas_inlet")
        assert len(iron_oc.fs.unit.gas_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_inlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_inlet")
        assert len(iron_oc.fs.unit.solid_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_inlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_inlet.temperature, Var)

        assert hasattr(iron_oc.fs.unit, "gas_outlet")
        assert len(iron_oc.fs.unit.gas_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_outlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "solid_outlet")
        assert len(iron_oc.fs.unit.solid_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.solid_outlet.flow_mass, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.particle_porosity, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.mass_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.solid_outlet.temperature, Var)

        assert isinstance(iron_oc.fs.unit.orifice_area, Constraint)
        assert isinstance(iron_oc.fs.unit.bed_area_eqn, Constraint)
        assert isinstance(iron_oc.fs.unit.bubble_area, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_emulsion_area, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_emulsion_area, Constraint)
        assert isinstance(iron_oc.fs.unit.bubble_cloud_heat_trans_coeff, Constraint)
        assert isinstance(iron_oc.fs.unit.convective_heat_trans_coeff, Constraint)
        assert isinstance(iron_oc.fs.unit.convective_heat_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.bubble_cloud_bulk_heat_trans, Constraint)
        assert isinstance(iron_oc.fs.unit.bubble_mass_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_emulsion_mass_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.bubble_heat_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_emulsion_heat_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_emulsion_heat_transfer, Constraint)

        assert number_variables(iron_oc) == 2059
        assert number_total_constraints(iron_oc) == 2014
        assert number_unused_variables(iron_oc) == 19

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        BFB = iron_oc.fs.unit  # alias to keep test lines short
        iscale.calculate_scaling_factors(BFB)
        initialization_tester(
            iron_oc,
            optarg={"tol": 1e-6},
            gas_phase_state_args={
                "flow_mol": 272.81,
                "temperature": 1186,
                "pressure": 1.86e5,
            },
            solid_phase_state_args={"flow_mass": 1230, "temperature": 1186},
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iron_oc):
        results = solver.solve(iron_oc, tee=True)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iron_oc):
        assert (
            pytest.approx(0.13964, abs=1e-5)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 0].value
        )
        assert (
            pytest.approx(1.09375, abs=1e-5)
            == iron_oc.fs.unit.velocity_superficial_gas[0, 1].value
        )
        assert (
            pytest.approx(0.01522, abs=1e-5)
            == iron_oc.fs.unit.bubble_diameter[0, 0].value
        )
        assert (
            pytest.approx(1.07748, abs=1e-5)
            == iron_oc.fs.unit.bubble_diameter[0, 1].value
        )
        assert (
            pytest.approx(0.37470, abs=1e-5)
            == iron_oc.fs.unit.velocity_bubble[0, 0].value
        )
        assert (
            pytest.approx(3.36531, abs=1e-5)
            == iron_oc.fs.unit.velocity_bubble[0, 1].value
        )
        assert pytest.approx(0.266906, abs=1e-5) == iron_oc.fs.unit.delta[0, 0].value
        assert pytest.approx(0.31323, abs=1e-5) == iron_oc.fs.unit.delta[0, 1].value
        assert (
            pytest.approx(126040.59980, rel=1e-4)
            == iron_oc.fs.unit.gas_outlet.pressure[0].value
        )
        assert (
            pytest.approx(59959.40019, rel=1e-4)
            == iron_oc.fs.unit.gas_inlet.pressure[0].value
            - iron_oc.fs.unit.gas_outlet.pressure[0].value
        )

    @pytest.mark.component
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iron_oc):
        # Conservation of material check
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_inlet_block[0].mw,
            iron_oc.fs.unit.gas_inlet_block[0].mw_eqn,
        )
        calculate_variable_from_constraint(
            iron_oc.fs.unit.gas_outlet_block[0].mw,
            iron_oc.fs.unit.gas_outlet_block[0].mw_eqn,
        )
        mbal_gas = value(
            (
                iron_oc.fs.unit.gas_inlet.flow_mol[0]
                * iron_oc.fs.unit.gas_inlet_block[0].mw
            )
            - (
                iron_oc.fs.unit.gas_outlet.flow_mol[0]
                * iron_oc.fs.unit.gas_outlet_block[0].mw
            )
        )
        mbal_solid = value(
            iron_oc.fs.unit.solid_inlet.flow_mass[0]
            - iron_oc.fs.unit.solid_outlet.flow_mass[0]
        )
        mbal_tol = mbal_gas + mbal_solid
        assert abs(mbal_tol) <= 1e-2

        # Reaction stoichiometric ratio check
        # Overall reducer reactions for methane combustion:
        # CH4 + 12Fe2O3 => 8Fe3O4 + CO2 + 2H2O
        mole_gas_reacted = value(
            iron_oc.fs.unit.gas_inlet.flow_mol[0]
            * iron_oc.fs.unit.gas_inlet.mole_frac_comp[0, "CH4"]
            - iron_oc.fs.unit.gas_outlet.flow_mol[0]
            * iron_oc.fs.unit.gas_outlet.mole_frac_comp[0, "CH4"]
        )
        mole_solid_reacted = value(
            (
                iron_oc.fs.unit.solid_inlet.flow_mass[0]
                * iron_oc.fs.unit.solid_inlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_inlet_block[0]._params.mw_comp["Fe2O3"]
            )
            - (
                iron_oc.fs.unit.solid_outlet.flow_mass[0]
                * iron_oc.fs.unit.solid_outlet.mass_frac_comp[0, "Fe2O3"]
                / iron_oc.fs.unit.solid_outlet_block[0]._params.mw_comp["Fe2O3"]
            )
        )
        stoichiometric_ratio = mole_solid_reacted / mole_gas_reacted
        assert pytest.approx(12, abs=1e-6) == stoichiometric_ratio

        # Conservation of energy check
        ebal_gas = value(
            (
                iron_oc.fs.unit.gas_inlet.flow_mol[0]
                * iron_oc.fs.unit.gas_inlet_block[0].enth_mol
            )
            - (
                iron_oc.fs.unit.gas_outlet.flow_mol[0]
                * iron_oc.fs.unit.gas_outlet_block[0].enth_mol
            )
        )
        ebal_solid = value(
            (
                iron_oc.fs.unit.solid_inlet.flow_mass[0]
                * iron_oc.fs.unit.solid_inlet_block[0].enth_mass
            )
            - (
                iron_oc.fs.unit.solid_outlet.flow_mass[0]
                * iron_oc.fs.unit.solid_outlet_block[0].enth_mass
            )
        )
        e_reaction = value(
            mole_gas_reacted
            * iron_oc.fs.unit.solid_emulsion.reactions[0, 0]._params.dh_rxn["R1"]
        )
        ebal_tol = ebal_gas + ebal_solid - e_reaction
        assert abs(ebal_tol) <= 1e-2

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        iron_oc.fs.unit.report()

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        iron_oc.fs.unit.gas_outlet.flow_mol[0].fix(1)

        with pytest.raises(InitializationError):
            iron_oc.fs.unit.initialize()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from abc import abstractmethod
from copy import deepcopy
import numpy as np

from ..transform_func import (
    TransformFunc,
    ShiftFunc,
    ScaleFunc,
    RotateFunc,
    ReflectFunc,
)


class Lattice(object):
    """A class used to represent crystal lattice locations.

    The class encodes methods for determining which Cartesian coordinates to
    consider as sites on an infinite crystal lattice. A ``Lattice`` can be constructed from
    a point on the lattice (i.e., a shift from the origin), an alignment (i.e., rotation from a
    nominal orientation), and appropriate scaling factors. With these attributes, we generally
    support the translation, rotation, and rescaling of lattices. Additionally, ``Lattice`` objects
    include a method for determining which sites should be considered neighbors.
    """

    # === DEFAULT CONSTRUCTOR
    def __init__(self):
        self._TransformFuncs = []

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        if isinstance(TransF, TransformFunc):
            self._TransformFuncs.append(TransF)
        else:
            raise TypeError

    def shift(self, Shift):
        if type(Shift) is ShiftFunc:
            self.applyTransF(Shift)
        elif type(Shift) is np.ndarray:
            self.applyTransF(ShiftFunc(Shift))
        else:
            raise TypeError

    def scale(self, Scale, OriginOfScale=None):
        if type(Scale) is ScaleFunc:
            self.applyTransF(Scale)
        elif type(Scale) is np.ndarray:
            self.applyTransF(ScaleFunc(Scale, OriginOfScale))
        else:
            raise TypeError

    def rotate(self, Rotation, OriginOfRotation=None):
        if type(Rotation) is RotateFunc:
            self.applyTransF(Rotation)
        elif type(Rotation) is np.ndarray:
            self.applyTransF(RotateFunc(Rotation, OriginOfRotation))
        else:
            raise TypeError

    def reflect(self, Reflection):
        if type(Reflection) is ReflectFunc:
            self.applyTransF(Reflection)
        else:
            raise TypeError

    # === PROPERTY EVALUATION METHODS
    @abstractmethod
    def isOnLattice(self, P):
        raise NotImplementedError

    @abstractmethod
    def areNeighbors(self, P1, P2):
        raise NotImplementedError

    @abstractmethod
    def getNeighbors(self, P, layer):
        raise NotImplementedError

    def _convertFromReference(self, P):
        for TransF in self._TransformFuncs:
            TransF.transform(P)

    def _convertToReference(self, P):
        for TransF in reversed(self._TransformFuncs):
            TransF.undo(P)

    def _getConvertFromReference(self, P):
        result = deepcopy(P)
        self._convertFromReference(result)
        return result

    def _getConvertToReference(self, P):
        result = deepcopy(P)
        self._convertToReference(result)
        return result



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Generic IDAES 1D Heat Exchanger Model with overall area and heat transfer coefficient
"""
# Import Pyomo libraries
from pyomo.environ import (
    Block,
    Var,
    check_optimal_termination,
    value,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool

# Import IDAES cores
from idaes.core import (
    ControlVolume1DBlock,
    UnitModelBlockData,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    FlowDirection,
    useDefault,
)
from idaes.models.unit_models.heat_exchanger import (
    HeatExchangerFlowPattern,
    hx_process_config,
    add_hx_references,
)
from idaes.core.util.config import is_physical_parameter_block, DefaultBool
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import ConfigurationError, InitializationError
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.initialization import SingleControlVolumeUnitInitializer
import idaes.logger as idaeslog


__author__ = "Jaffer Ghouse, Andrew Lee"

# Set up logger
_log = idaeslog.getLogger(__name__)


class HX1DInitializer(SingleControlVolumeUnitInitializer):
    """
    Initializer for 1D Heat Exchanger units.

    """

    def initialization_routine(
        self,
        model: Block,
        plugin_initializer_args: dict = None,
        duty=None,
    ):
        """
        Common initialization routine for 1D Heat Exchangers.

        This routine starts by initializing the hot and cold side properties. Next, the heat
        transfer between the two sides is fixed to an initial guess for the heat duty (provided by the duty
        argument), the associated constraints is deactivated, and the model is then solved. Finally, the heat
        duty is unfixed and the heat transfer constraints reactivated followed by a final solve of the model.

        Args:
            model: Pyomo Block to be initialized
            plugin_initializer_args: dict-of-dicts containing arguments to be passed to plug-in Initializers.
                Keys should be submodel components.
            duty: initial guess for heat duty to assist with initialization. Can be a Pyomo expression with units.
                Default is 1/4*U*A*(Thot - Tcold)

        Returns:
            Pyomo solver results object
        """
        return super(SingleControlVolumeUnitInitializer, self).initialization_routine(
            model=model,
            plugin_initializer_args=plugin_initializer_args,
            duty=duty,
        )

    def initialize_main_model(
        self,
        model: Block,
        duty=None,
    ):
        """
        Initialization routine for main 1D HX models.

        Args:
            model: Pyomo Block to be initialized.
            duty: initial guess for heat duty to assist with initialization. Can be a Pyomo expression with units.

        Returns:
            Pyomo solver results object.

        """
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        # Create solver
        solver = self._get_solver()

        # ---------------------------------------------------------------------
        # Get length values
        if model.length.fixed:
            # Most likely case
            model.cold_side.length.set_value(model.length)
        elif model.cold_side.length.fixed:
            # This would be unusual, but check
            model.length.set_value(model.cold_side.length)
        else:
            # No fixed value for length - we will assume the user knows what they are doing
            pass

        # Initialize control volumes blocks
        Lfix = model.hot_side.length.fixed
        model.hot_side.length.fix()
        self.initialize_control_volume(model.hot_side)

        if not Lfix:
            model.hot_side.length.unfix()

        Lfix = model.cold_side.length.fixed
        model.cold_side.length.fix()
        self.initialize_control_volume(model.cold_side)

        if not Lfix:
            model.cold_side.length.unfix()

        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit with fixed heat duty
        # Guess heat duty based on 1/4 of maximum driving force
        hot_side_units = (
            model.hot_side.config.property_package.get_metadata().get_derived_units
        )
        if duty is None:
            duty = (
                0.25
                * model.heat_transfer_coefficient[0, 0]
                * model.area
                * (
                    model.hot_side.properties[0, 0].temperature
                    - pyunits.convert(
                        model.cold_side.properties[0, 0].temperature,
                        to_units=hot_side_units("temperature"),
                    )
                )
            )

        duty_per_length = duty / model.length
        # Fix heat duties
        for v in model.hot_side.heat.values():
            v.fix(-duty_per_length)
        for v in model.cold_side.heat.values():
            v.fix(duty_per_length)

        # Deactivate heat duty constraints
        model.heat_transfer_eq.deactivate()
        model.heat_conservation.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)
        init_log.info_high(f"Initialization Step 2 {idaeslog.condition(res)}.")

        # Unfix heat duty and reactivate constraints
        for v in model.hot_side.heat.values():
            v.unfix()
        for v in model.cold_side.heat.values():
            v.unfix()
        model.heat_transfer_eq.activate()
        model.heat_conservation.activate()
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)
        init_log.info_high(f"Initialization Step 3 {idaeslog.condition(res)}.")

        return res


@declare_process_block_class("HeatExchanger1D")
class HeatExchanger1DData(UnitModelBlockData):
    """Standard Heat Exchanger 1D Unit Model Class."""

    default_initializer = HX1DInitializer

    CONFIG = UnitModelBlockData.CONFIG(implicit=True)
    # Template for config arguments for hot and cold side
    _SideTemplate = ConfigBlock()
    _SideTemplate.declare(
        "dynamic",
        ConfigValue(
            default=useDefault,
            domain=DefaultBool,
            description="Dynamic model flag",
            doc="""Indicates whether this model will be dynamic or not,
**default** = useDefault.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - set as a dynamic model,
**False** - set as a steady-state model.}""",
        ),
    )
    _SideTemplate.declare(
        "has_holdup",
        ConfigValue(
            default=useDefault,
            domain=DefaultBool,
            description="Holdup construction flag",
            doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - construct holdup terms,
**False** - do not construct holdup terms}""",
        ),
    )
    _SideTemplate.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    _SideTemplate.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    _SideTemplate.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should
be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    _SideTemplate.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    _SideTemplate.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Phase equilibrium term construction flag",
            doc="""Argument to enable phase equilibrium.
- True - include phase equilibrium term
- False - do not include phase equilibrium term""",
        ),
    )
    _SideTemplate.declare(
        "property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property
calculations
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a ParameterBlock object""",
        ),
    )
    _SideTemplate.declare(
        "property_package_args",
        ConfigValue(
            default={},
            description="Arguments for constructing property package",
            doc="""A dict of arguments to be passed to the PropertyBlockData
and used when constructing these
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a dict (see property package for documentation)""",
        ),
    )
    # TODO : We should probably think about adding a consistency check for the
    # TODO : discretisation methods as well.
    _SideTemplate.declare(
        "transformation_method",
        ConfigValue(
            default=useDefault,
            description="Discretization method to use for DAE transformation",
            doc="""Discretization method to use for DAE transformation. See
Pyomo documentation for supported transformations.""",
        ),
    )
    _SideTemplate.declare(
        "transformation_scheme",
        ConfigValue(
            default=useDefault,
            description="Discretization scheme to use for DAE transformation",
            doc="""Discretization scheme to use when transformating domain. See
Pyomo documentation for supported schemes.""",
        ),
    )

    # Create individual config blocks for hot and cold side
    CONFIG.declare("hot_side", _SideTemplate(doc="hot side config arguments"))
    CONFIG.declare("cold_side", _SideTemplate(doc="cold side config arguments"))

    # Common config args for both sides
    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=20,
            domain=int,
            description="Number of finite elements length domain",
            doc="""Number of finite elements to use when discretizing length
domain (default=20)""",
        ),
    )
    CONFIG.declare(
        "collocation_points",
        ConfigValue(
            default=5,
            domain=int,
            description="Number of collocation points per finite element",
            doc="""Number of collocation points to use per finite element when
discretizing length domain (default=3)""",
        ),
    )
    CONFIG.declare(
        "flow_type",
        ConfigValue(
            default=HeatExchangerFlowPattern.cocurrent,
            domain=In(HeatExchangerFlowPattern),
            description="Flow configuration of heat exchanger",
            doc="""Flow configuration of heat exchanger
- HeatExchangerFlowPattern.cocurrent: hot and cold flows from 0 to 1
(default)
- HeatExchangerFlowPattern.countercurrent: hot side flows from 0 to 1
cold side flows from 1 to 0""",
        ),
    )
    CONFIG.declare(
        "hot_side_name",
        ConfigValue(
            default=None,
            domain=str,
            doc="Hot side name, sets control volume and inlet and outlet names. "
            "Default = None.",
        ),
    )
    CONFIG.declare(
        "cold_side_name",
        ConfigValue(
            default=None,
            domain=str,
            doc="Cold side name, sets control volume and inlet and outlet names. "
            "Default = None.",
        ),
    )

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super().build()
        self._process_config()

        # Set flow directions for the control volume blocks and specify
        # discretization if not specified.
        if self.config.hot_side.transformation_method is useDefault:
            _log.info(
                "Discretization method was "
                "not specified for the hot side of the "
                "heat exchanger. "
                "Defaulting to finite "
                "difference method on the hot side."
            )
            self.config.hot_side.transformation_method = "dae.finite_difference"
        if self.config.cold_side.transformation_method is useDefault:
            _log.info(
                "Discretization method was "
                "not specified for the cold side of the "
                "heat exchanger. "
                "Defaulting to finite "
                "difference method on the cold side."
            )
            self.config.cold_side.transformation_method = "dae.finite_difference"

        if (
            self.config.hot_side.transformation_method
            != self.config.cold_side.transformation_method
        ):
            raise ConfigurationError(
                "HeatExchanger1D only supports the same transformation "
                "method on the hot and cold side domains for "
                "both cocurrent and countercurrent flow patterns. "
                f"Found method {self.config.hot_side.transformation_method} "
                f"on hot side and method {self.config.cold_side.transformation_method} "
                "on cold side."
            )
        if self.config.hot_side.transformation_method == "dae.collocation":
            if (
                self.config.hot_side.transformation_scheme is useDefault
                or self.config.cold_side.transformation_scheme is useDefault
            ):
                raise ConfigurationError(
                    "If a collocation method is used for HeatExchanger1D, the user "
                    "must specify the transformation scheme they want to use."
                )
            if (
                self.config.hot_side.transformation_scheme
                != self.config.cold_side.transformation_scheme
            ):
                raise ConfigurationError(
                    "If a collocation method is used, "
                    "HeatExchanger1D only supports the same transformation "
                    "scheme on the hot and cold side domains. Found "
                    f"{self.config.hot_side.transformation_scheme} scheme on "
                    f"hot side and {self.config.cold_side.transformation_scheme} "
                    "scheme on cold side."
                )
        else:
            if (
                self.config.hot_side.transformation_scheme
                != self.config.cold_side.transformation_scheme
            ):
                _log.caution(
                    "The hot and cold sides are being discretized with different "
                    "discretization schemes. While this may result in better numerical "
                    "stability if an upwind scheme is used, it may also result in "
                    "energy conservation errors. High-order collocation methods can "
                    "provide both accuracy and numerical stability."
                )

        if self.config.hot_side.transformation_scheme is useDefault:
            _log.info(
                "Discretization scheme was "
                "not specified for the hot side of the "
                "counter-current heat exchanger. "
                "Defaulting to backward finite "
                "difference on the hot side."
            )
            self.config.hot_side.transformation_scheme = "BACKWARD"
        if self.config.cold_side.transformation_scheme is useDefault:
            _log.info(
                "Discretization scheme was "
                "not specified for the cold side of the "
                "counter-current heat exchanger. "
                "Defaulting to backward finite "
                "difference on the cold side."
            )
            self.config.cold_side.transformation_scheme = "BACKWARD"

        if self.config.flow_type == HeatExchangerFlowPattern.cocurrent:
            set_direction_hot = FlowDirection.forward
            set_direction_cold = FlowDirection.forward
            if self.config.hot_side.transformation_scheme != "BACKWARD":
                _log_upwinding_disclaimer(
                    "hot_side",
                    "cocurrent",
                    self.config.hot_side.transformation_scheme,
                    "BACKWARD",
                )
            if self.config.cold_side.transformation_scheme != "BACKWARD":
                _log_upwinding_disclaimer(
                    "cold_side",
                    "cocurrent",
                    self.config.cold_side.transformation_scheme,
                    "BACKWARD",
                )

        elif self.config.flow_type == HeatExchangerFlowPattern.countercurrent:
            set_direction_hot = FlowDirection.forward
            set_direction_cold = FlowDirection.backward
            if self.config.hot_side.transformation_scheme != "BACKWARD":
                _log_upwinding_disclaimer(
                    "hot_side",
                    "countercurrent",
                    self.config.hot_side.transformation_scheme,
                    "BACKWARD",
                )
            if self.config.cold_side.transformation_scheme != "FORWARD":
                _log_upwinding_disclaimer(
                    "cold_side",
                    "countercurrent",
                    self.config.cold_side.transformation_scheme,
                    "FORWARD",
                )
        else:
            raise ConfigurationError(
                "{} HeatExchanger1D only supports cocurrent and "
                "countercurrent flow patterns, but flow_type configuration"
                " argument was set to {}.".format(self.name, self.config.flow_type)
            )

        # Control volume 1D for hot
        self.hot_side = ControlVolume1DBlock(
            dynamic=self.config.hot_side.dynamic,
            has_holdup=self.config.hot_side.has_holdup,
            property_package=self.config.hot_side.property_package,
            property_package_args=self.config.hot_side.property_package_args,
            transformation_method=self.config.hot_side.transformation_method,
            transformation_scheme=self.config.hot_side.transformation_scheme,
            finite_elements=self.config.finite_elements,
            collocation_points=self.config.collocation_points,
        )

        self.cold_side = ControlVolume1DBlock(
            dynamic=self.config.cold_side.dynamic,
            has_holdup=self.config.cold_side.has_holdup,
            property_package=self.config.cold_side.property_package,
            property_package_args=self.config.cold_side.property_package_args,
            transformation_method=self.config.cold_side.transformation_method,
            transformation_scheme=self.config.cold_side.transformation_scheme,
            finite_elements=self.config.finite_elements,
            collocation_points=self.config.collocation_points,
        )

        self.hot_side.add_geometry(flow_direction=set_direction_hot)
        self.cold_side.add_geometry(flow_direction=set_direction_cold)

        self.hot_side.add_state_blocks(
            information_flow=set_direction_hot,
            has_phase_equilibrium=self.config.hot_side.has_phase_equilibrium,
        )
        self.cold_side.add_state_blocks(
            information_flow=set_direction_cold,
            has_phase_equilibrium=self.config.cold_side.has_phase_equilibrium,
        )

        # Populate hot
        self.hot_side.add_material_balances(
            balance_type=self.config.hot_side.material_balance_type,
            has_phase_equilibrium=self.config.hot_side.has_phase_equilibrium,
        )

        self.hot_side.add_energy_balances(
            balance_type=self.config.hot_side.energy_balance_type,
            has_heat_transfer=True,
        )

        self.hot_side.add_momentum_balances(
            balance_type=self.config.hot_side.momentum_balance_type,
            has_pressure_change=self.config.hot_side.has_pressure_change,
        )

        self.hot_side.apply_transformation()

        # Populate cold side
        self.cold_side.add_material_balances(
            balance_type=self.config.cold_side.material_balance_type,
            has_phase_equilibrium=self.config.cold_side.has_phase_equilibrium,
        )

        self.cold_side.add_energy_balances(
            balance_type=self.config.cold_side.energy_balance_type,
            has_heat_transfer=True,
        )

        self.cold_side.add_momentum_balances(
            balance_type=self.config.cold_side.momentum_balance_type,
            has_pressure_change=self.config.cold_side.has_pressure_change,
        )

        self.cold_side.apply_transformation()

        # Add Ports for hot side
        self.add_inlet_port(name="hot_side_inlet", block=self.hot_side)
        self.add_outlet_port(name="hot_side_outlet", block=self.hot_side)

        # Add Ports for cold side
        self.add_inlet_port(name="cold_side_inlet", block=self.cold_side)
        self.add_outlet_port(name="cold_side_outlet", block=self.cold_side)

        # Add references to the user provided aliases if applicable
        add_hx_references(self)

        self._make_geometry()
        self._make_performance()

        hot_side_units = (
            self.hot_side.config.property_package.get_metadata().get_derived_units
        )
        q_units = hot_side_units("power") / hot_side_units("length")

        @self.Constraint(
            self.flowsheet().time,
            self.hot_side.length_domain,
            doc="Heat conservation equality",
        )
        def heat_conservation(self, t, x):
            return pyunits.convert(self.cold_side.heat[t, x], to_units=q_units) == -(
                self.hot_side.heat[t, x]
            )

    def _process_config(self):
        hx_process_config(self)

    def _make_geometry(self):
        hot_side_units = (
            self.hot_side.config.property_package.get_metadata().get_derived_units
        )

        self.area = Var(
            initialize=1, units=hot_side_units("area"), doc="Heat transfer area"
        )

        # Add reference to control volume geometry
        add_object_reference(self, "length", self.hot_side.length)

        # Equate hot and cold side lengths
        @self.Constraint(doc="Equating hot and cold side lengths")
        def length_equality(self):
            return (
                pyunits.convert(
                    self.cold_side.length, to_units=hot_side_units("length")
                )
                == self.hot_side.length
            )

    def _make_performance(self):
        """
        Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        hot_side_units = (
            self.hot_side.config.property_package.get_metadata().get_derived_units
        )

        # Performance variables
        self.heat_transfer_coefficient = Var(
            self.flowsheet().time,
            self.hot_side.length_domain,
            initialize=50,
            doc="Average heat transfer coefficient",
            units=hot_side_units("heat_transfer_coefficient"),
        )

        @self.Constraint(
            self.flowsheet().time,
            self.hot_side.length_domain,
            doc="Heat transfer between hot_side and cold_side",
        )
        def heat_transfer_eq(self, t, x):
            return self.hot_side.heat[t, x] == -(
                self.heat_transfer_coefficient[t, x]
                * self.area
                / self.length
                * (
                    self.hot_side.properties[t, x].temperature
                    - pyunits.convert(
                        self.cold_side.properties[t, x].temperature,
                        to_units=hot_side_units("temperature"),
                    )
                )
            )

    def initialize_build(
        self,
        hot_side_state_args=None,
        cold_side_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        duty=None,
    ):
        """
        Initialization routine for the unit.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            duty : an initial guess for the amount of heat transferred. This
                should be a tuple in the form (value, units). A default value
                is calculated based on stream temperatures, the overall
                heat transfer coefficient, and exchanger area

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Get length values
        if self.length.fixed:
            # Most likely case
            self.cold_side.length.set_value(self.length)
        elif self.cold_side.length.fixed:
            # This would be unusual, but check
            self.length.set_value(self.cold_side.length)
        else:
            # No fixed value for length - we will assume the user knows what they are doing
            pass

        # Initialize control volumes blocks
        Lfix = self.hot_side.length.fixed
        self.hot_side.length.fix()
        flags_hot_side = self.hot_side.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=hot_side_state_args,
        )
        if not Lfix:
            self.hot_side.length.unfix()

        Lfix = self.cold_side.length.fixed
        self.cold_side.length.fix()
        flags_cold_side = self.cold_side.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=cold_side_state_args,
        )
        if not Lfix:
            self.cold_side.length.unfix()

        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit with fixed heat duty
        # Guess heat duty based on 1/4 of maximum driving force
        hot_side_units = (
            self.hot_side.config.property_package.get_metadata().get_derived_units
        )
        cold_side_units = (
            self.cold_side.config.property_package.get_metadata().get_derived_units
        )
        t0 = self.flowsheet().time.first()
        if duty is None:
            duty = value(
                0.25
                * self.heat_transfer_coefficient[t0, 0]
                * self.area
                * (
                    self.hot_side.properties[t0, 0].temperature
                    - pyunits.convert(
                        self.cold_side.properties[t0, 0].temperature,
                        to_units=hot_side_units("temperature"),
                    )
                )
            )
        else:
            duty = pyunits.convert_value(
                duty[0], from_units=duty[1], to_units=hot_side_units("power")
            )
        duty_per_length = value(duty / self.length)
        # Fix heat duties
        for v in self.hot_side.heat.values():
            v.fix(-duty_per_length)
        for v in self.cold_side.heat.values():
            v.fix(
                pyunits.convert_value(
                    duty_per_length,
                    to_units=cold_side_units("power") / cold_side_units("length"),
                    from_units=hot_side_units("power") / hot_side_units("length"),
                )
            )

        # Deactivate heat duty constraints
        self.heat_transfer_eq.deactivate()
        self.heat_conservation.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # Unfix heat duty and reactivate constraints
        for v in self.hot_side.heat.values():
            v.unfix()
        for v in self.cold_side.heat.values():
            v.unfix()
        self.heat_transfer_eq.activate()
        self.heat_conservation.activate()
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        self.hot_side.release_state(flags_hot_side)
        self.cold_side.release_state(flags_cold_side)

        if res is not None and not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        init_log.info("Initialization Complete.")

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Area"] = self.area
        var_dict["Length"] = self.length

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        # Get names for hot and cold sides
        hot_name = self.config.hot_side_name
        if hot_name is None:
            hot_name = "Hot Side"
        cold_name = self.config.cold_side_name
        if cold_name is None:
            cold_name = "Cold Side"
        return create_stream_table_dataframe(
            {
                f"{hot_name} Inlet": self.hot_side_inlet,
                f"{hot_name} Outlet": self.hot_side_outlet,
                f"{cold_name} Inlet": self.cold_side_inlet,
                f"{cold_name} Outlet": self.cold_side_outlet,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for i, c in self.heat_transfer_eq.items():
            iscale.constraint_scaling_transform(
                c,
                iscale.get_scaling_factor(
                    self.hot_side.heat[i], default=1, warning=True
                ),
                overwrite=False,
            )

        for i, c in self.heat_conservation.items():
            iscale.constraint_scaling_transform(
                c,
                iscale.get_scaling_factor(
                    self.hot_side.heat[i], default=1, warning=True
                ),
                overwrite=False,
            )


def _log_upwinding_disclaimer(side, flow_config, bad_scheme, good_scheme):
    _log.info(
        f"For {side}, a {bad_scheme} scheme was chosen to discretize the length domain. "
        f"However, this scheme is not an upwind scheme for {flow_config} flow, and "
        f"as a result may run into numerical stability issues. To avoid this, "
        f"use a {good_scheme} scheme (which may result in energy conservation issues "
        "for coarse discretizations) or use a high-order collocation method."
    )
